/*
 * e-import-assistant.h
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with the program; if not, see <http://www.gnu.org/licenses/>
 *
 *
 * Copyright (C) 1999-2008 Novell, Inc. (www.novell.com)
 *
 */

#ifndef E_IMPORT_ASSISTANT_H
#define E_IMPORT_ASSISTANT_H

#include <gtk/gtk.h>

/* Standard GObject macros */
#define E_TYPE_IMPORT_ASSISTANT \
	(e_import_assistant_get_type ())
#define E_IMPORT_ASSISTANT(obj) \
	(G_TYPE_CHECK_INSTANCE_CAST \
	((obj), E_TYPE_IMPORT_ASSISTANT, EImportAssistant))
#define E_IMPORT_ASSISTANT_CLASS(cls) \
	(G_TYPE_CHECK_CLASS_CAST \
	((cls), E_TYPE_IMPORT_ASSISTANT, EImportAssistantClass))
#define E_IS_IMPORT_ASSISTANT(obj) \
	(G_TYPE_CHECK_INSTANCE_TYPE \
	((obj), E_TYPE_IMPORT_ASSISTANT))
#define E_IS_IMPORT_ASSISTANT_CLASS(cls) \
	(G_TYPE_CHECK_CLASS_TYPE \
	((cls), E_TYPE_IMPORT_ASSISTANT))
#define E_IMPORT_ASSISTANT_GET_CLASS(obj) \
	(G_TYPE_INSTANCE_GET_CLASS \
	((obj), E_TYPE_IMPORT_ASSISTANT, EImportAssistantClass))

G_BEGIN_DECLS

typedef struct _EImportAssistant EImportAssistant;
typedef struct _EImportAssistantClass EImportAssistantClass;
typedef struct _EImportAssistantPrivate EImportAssistantPrivate;

struct _EImportAssistant {
	GtkAssistant parent;
	EImportAssistantPrivate *priv;
};

struct _EImportAssistantClass {
	GtkAssistantClass parent_class;
};

GType		e_import_assistant_get_type	(void);
GtkWidget *	e_import_assistant_new		(GtkWindow *parent);
GtkWidget *	e_import_assistant_new_simple	(GtkWindow *parent,
						 gchar **uris);

G_END_DECLS

#endif /* E_IMPORT_ASSISTANT_H */
