/* Automatically generated. Do not edit. */
gchar *s = N_("Adjust Score");
gchar *s = N_("Assign Color");
gchar *s = N_("Assign Score");
gchar *s = N_("Attachments");
gchar *s = N_("BCC");
gchar *s = N_("Beep");
gchar *s = N_("CC");
gchar *s = N_("Completed On");
gchar *s = N_("contains");
gchar *s = N_("Copy to Folder");
gchar *s = N_("Date received");
gchar *s = N_("Date sent");
gchar *s = N_("Delete");
gchar *s = N_("Deleted");
gchar *s = N_("does not contain");
gchar *s = N_("does not end with");
gchar *s = N_("does not exist");
gchar *s = N_("does not return");
gchar *s = N_("does not sound like");
gchar *s = N_("does not start with");
gchar *s = N_("Do Not Exist");
gchar *s = N_("Draft");
gchar *s = N_("ends with");
gchar *s = N_("Exist");
gchar *s = N_("exists");
gchar *s = N_("Expression");
gchar *s = N_("Follow Up");
gchar *s = N_("Forward to");
gchar *s = N_("Important");
gchar *s = N_("is");
gchar *s = N_("is after");
gchar *s = N_("is before");
gchar *s = N_("is Flagged");
gchar *s = N_("is greater than");
gchar *s = N_("is less than");
gchar *s = N_("is not");
gchar *s = N_("is not Flagged");
gchar *s = N_("is not set");
gchar *s = N_("is set");
gchar *s = N_("Junk");
gchar *s = N_("Junk Test");
gchar *s = N_("Label");
gchar *s = N_("Mailing list");
gchar *s = N_("Match All");
gchar *s = N_("Message Body");
gchar *s = N_("Message Header");
gchar *s = N_("Message is Junk");
gchar *s = N_("Message is not Junk");
gchar *s = N_("Message Location");
gchar *s = N_("Move to Folder");
gchar *s = N_("Pipe to Program");
gchar *s = N_("Play Sound");
gchar *s = N_("Read");
gchar *s = N_("Recipients");
gchar *s = N_("Regex Match");
gchar *s = N_("Replied to");
gchar *s = N_("returns");
gchar *s = N_("returns greater than");
gchar *s = N_("returns less than");
gchar *s = N_("Run Program");
gchar *s = N_("Score");
gchar *s = N_("Sender");
gchar *s = N_("Sender or Recipients");
gchar *s = N_("Set Label");
gchar *s = N_("Set Status");
gchar *s = N_("Size (kB)");
gchar *s = N_("sounds like");
gchar *s = N_("Source Account");
gchar *s = N_("Specific header");
gchar *s = N_("starts with");
gchar *s = N_("Status");
gchar *s = N_("Stop Processing");
gchar *s = N_("Subject");
gchar *s = N_("Unset Status");
